import os, sys
from pkg_resources import resource_filename

from zicbee.core.zutils import DEBUG

import web
import web.wsgiserver
from zicbee.player.webplayer import webplayer, web_db_index

def do_serve(pure=False):
    """ Create a ZicDB instance
    parameters:
        pure (default: False): just start DB serving, no player
    """
    # chdir to serve files at the right place

    p = os.path.dirname(resource_filename('zicbee.ui.web', 'static'))
    os.chdir( p )

    # let's do webplayer

    sys.argv = ['zicdb', '0.0.0.0:9090']
    try:
        print "Running webplayer from", __file__
        if pure:
            urls = ('/db/(.*)', 'web_db_index',
                    '/(.*)', 'web_db_index')
        else:
            urls = ('/db/(.*)', 'web_db_index',
                    '/(.*)', 'webplayer')
        app = web.application(urls, locals())
        app.run()
    except:
        DEBUG()
        print os.kill(os.getpid(), 9)
        #print 'kill', os.getpid()




__all__ = ['do_get']

import os
import sys
import time
import urllib
from itertools import chain
from weakref import WeakKeyDictionary

from .search import do_search
from zicbee.core.zutils import duration_tidy
from zicbee.core.config import config

def DownloadGenerator(uri):
    uri, filename = uri

    if os.path.exists(filename):
        return

    site = urllib.urlopen(uri)
    out_file = file(filename, 'w')
    BUF_SZ = 2**16
    total_size = int(site.info().getheader('Content-Length'))
    actual_size = 0
    progress_p = 0

    while True:
        data = site.read(BUF_SZ)
        if not data:
            out_file.close()
            break
        out_file.write(data)
        actual_size += len(data)
        percent = total_size/actual_size
        if percent != progress_p:
            yield percent

            progress_p = percent
        else:
            yield '.'


class Downloader(object):
    def __init__(self, nb_dl=2):
        self._nb_dl = nb_dl
        self._last_display = time.time()

    def run(self, uri_list):
        downloaders = [] # Generators to handle
        in_queue = [] # List of "TODO" uris

        _download_infos = dict(count=0, start_ts = time.time())
        percent_memory = WeakKeyDictionary()
        write_out = sys.stdout.write

        def _download():
            for dl in downloaders:
                try:
                    ret = dl.next()
                except StopIteration:
                    downloaders.remove(dl)
                    _download_infos['count'] += 1
                else:
                    if isinstance(ret, int):
                        percent_memory[dl] = ret

            # Display things
            t = time.time()

            if self._last_display + 0.1 < t:
                self._last_display = t
                sumup = ', '.join('%3d%%'%(val if int(val)<=100 else 0)
                        for val in percent_memory.itervalues())
                write_out(' [ %s ] %d               \r'%(sumup, _download_infos['count']))

                sys.stdout.flush()

        for uri in chain( uri_list, in_queue ):
            if len(downloaders) < self._nb_dl:
                try:
                    dg = DownloadGenerator(uri)
                    dg.next() # start the pump
                    downloaders.append( dg )
                except StopIteration:
                    pass
            else:
                in_queue.append(uri) # Later please !
                # iterate
                _download()

        # Terminate the job
        while downloaders:
            _download()

        t = time.time() - _download_infos['start_ts']
        write_out("                         \nGot %d files in %s. Enjoy ;)\n"%(
                _download_infos['count'], duration_tidy(t)))


def do_get(host=config.db_host, out=config.download_dir):
    """ Get songs (same syntax as search)
    See "help" for a more complete documentation
    """
    if ':' not in host:
        host += ':9090'

    uri_list = []
    def _append_uri_filename(args):
        uri = args[0]
        filename = os.path.join(out,
                ' - '.join(a for a in args[1:4] if a)
                + args[0].split('?', 1)[0][-4:])
        uri_list.append((uri, filename))

    do_search(out=_append_uri_filename, host=host)
    Downloader().run(uri_list)




# vim: et ts=4 sw=4
from zicbee.db import valid_tags

def do_help():
        """ Show a nifty help about most used commands """
        print "Welcome to ZicDB!".center(80)
        print """
use
    Not a command by itself, used to specify active database (default: songs)
    Exemple:
    %% %(prog)s use lisa search artist: dire st

serve[::pure]
    Runs a user-accessible www server on port 8080

  pure:
    don't allow player functionality access

list
    List available Databases.

reset
    Erases the Database (every previous scan is lost!)

bundle <filename>
    Create a bundle (compressed archive) of the database

scan <directory|archive> [directory|archive...]
    Scan directories/archive for files and add them to the database

get[::host][::out] <match command>
  host:
    the host to connect to

  out:
    the output directory

  Example:
    %% %(prog)s get::out=/tmp/download::host=gunter.biz artist: black
    %% %(prog)s get::gunter.biz artist: black

search[::out] <match command>

  out:
    specifies the output format (for now: m3u or null or default)

--- Match commands composition: ---

    field: value [ [or|and] field2: value2 ]...
    for length, value may be preceded by "<" or ">"
    if field name starts with a capital, the search is case-sensitive

  Possible fields:
\t- %(tags)s

  Exemple:
  %% %(prog)s search filename: shak length > 3*60

fullhelp
    List all available commands (with associated help if available)
    """%dict(
            tags = '\n\t- '.join(valid_tags),
            prog = "zicdb")




#!/usr/bin/env python

try:
    import pygtk
    pygtk.require("2.0")
finally:
    try:
        import gtk
        import gtk.glade
    except:
        raise SystemExit("Can't load gtk+")

import gobject
import sys
from pkg_resources import resource_filename
from zicbee.core.config import config
from zicbee.core.zutils import duration_tidy, DEBUG
from zicbee.player.events import DelayedAction, IterableAction
from zicbee.player.playerlogic import PlayerCtl

class PPlayer(object):

    def __init__(self, player_ctl):
        self.player_ctl = player_ctl

        self._old_size = (4, 4)
        self._actual_infos = ''

        self._wtree = gtk.glade.XML(resource_filename('zicbee.ui.gtk', 'player.glade'))

        self.win = self._wtree.get_widget('main_window')
        self.list_w = self._wtree.get_widget('songlist_tv')

        self.list_w.set_model(self.player_ctl.playlist)
        for i, name in enumerate(('Artist', 'Album', 'Title')):
            col = gtk.TreeViewColumn(name, gtk.CellRendererText(), text=i+1)
            col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
            col.set_fixed_width(120)
            col.set_resizable(True)
            self.list_w.append_column( col )

        self.pat = self._wtree.get_widget('pattern_entry')
        self.pat.set_text(config.default_search)
        self.pat.grab_focus()
        self.info_lbl = self._wtree.get_widget('info_label')
        self.time_lbl = self._wtree.get_widget('elapsed_time')
        # position
        self.cursor = self._wtree.get_widget('cursor')
        self.cursor.set_range(0, 100)
        self.cursor.set_show_fill_level(True)
        self.cursor.set_fill_level(0)
        # uri
        self._song_uri_lbl = self._wtree.get_widget('song_uri')
        # volume
        self.volume_w = self._wtree.get_widget('volume')
        self.volume_w.set_value(100)

        self.hostname_w = self._wtree.get_widget('hostname')
        if len(sys.argv) == 2:
            self.hostname_w.set_text(sys.argv[1])
        else:
            self.hostname_w.set_text(config.db_host)

        self.status_w = self._wtree.get_widget('statusbar')
        self.status_w_ctx = self.status_w.get_context_id('player')
        self._push_status('idle')

        handlers = dict( (prop, getattr(self, prop))
            for prop in dir(self)
            if prop[0] != '_' and callable(getattr(self, prop))
            )
        self._wtree.signal_autoconnect(handlers)
        self.win.connect('destroy', gtk.main_quit)
        self.win.set_geometry_hints(self.win, 386, 176)
        self.win.show()
        self._old_size = self.win.get_size()
        self._popup_win = None

    def _push_status(self, txt):
        if txt is None:
            self._pop_status()
        else:
            self.status_w.push(self.status_w_ctx, txt)

    def _pop_status(self):
        self.status_w.pop(self.status_w_ctx)

    def change_volume(self, w, value):
        value *= 100
        if not (0 <= value <= 100):
            return False
        self.player_ctl._volume_action.start(args=int(value))

    def absolute_seek(self, w, type, val):
        #     |      Seek to some place in the movie.
        #     |          0 is a relative seek of +/- <value> seconds (default).
        #     |          1 is a seek to <value> % in the movie.
        #     |          2 is a seek to an absolute position of <value> seconds.
        self.player_ctl.seek(val)

    def validate_pattern(self, w):
        hostname = self.hostname_w.get_text()
        try:
            pat = self.pat.get_text()
            it = self.player_ctl.fetch_playlist(hostname, pattern=pat)
            it.next()
            IterableAction(it).start(0.001, prio=gobject.PRIORITY_DEFAULT_IDLE)
        except:
            DEBUG()
            self._pop_status()
            self._push_status('Connection to %s failed'%hostname)
        else:
            config.default_search = pat
            self._pop_status()
            self._push_status('Connected' if len(self.player_ctl.playlist) else 'Empty')

        DelayedAction(self.player_ctl._play_selected, 0).start(0.5)

    def shuffle_playlist(self, w):
        self.player_ctl.shuffle()

    def toggle_pause(self, w):
        self.player_ctl.pause()

    def play_prev(self, w):
        try:
            self.player_ctl.select(-1)
        except IndexError:
            pass

    def play_next(self, *args):
        try:
            self.player_ctl.select(1)
        except IndexError:
            pass

    def toggle_playlist(self, expander):
        DelayedAction(self.win.resize, *self._old_size).start(0.3)
        self._old_size = self.win.get_size()

    def force_change_song(self, treeview, path, treeview_col):
        self.player_ctl._play_selected(path[0])

    def SIG_update_infos(self, infos):
        # XXX: refactor this
        self.info_lbl.set_markup('\n'.join(infos))

    def SIG_update_total(self, total):
        self._actual_infos = duration_tidy(total)
        self._pop_status()
        self._push_status('%d songs | %s'%(len(self.player_ctl.playlist), self._actual_infos))

    def SIG_status_changed(self, status):
        if status == 'stopped':
            self.info_lbl.set_text('Not Playing.')

    def SIG_progress(self, val):
        self.cursor.set_value(val)
        self.time_lbl.set_text(duration_tidy(val))

    def SIG_download_progress(self, val):
        self.cursor.set_fill_level(val)

    def SIG_song_length(self, length):
        self.cursor.set_range(0, length)

    def SIG_song_uri(self, uri):
        self._song_uri_lbl.set_text(uri)

    def SIG_select(self, pos):
        self.list_w.set_cursor( (pos, 0) )
        from pyglet import font
        from pyglet import window
        WIDTH = 600
        HEIGHT = 100

        if self._popup_win is None:
            self._popup_win = window.Window(WIDTH, HEIGHT, visible=False, style=window.Window.WINDOW_STYLE_TOOL)
        win = self._popup_win
        win.switch_to()
        # screen infos:
        screen0 = window.get_platform().get_default_display().get_screens()[0]

        # txt
        ft = font.load('Arial', HEIGHT/4)
        txt = self.info_lbl.get_text().decode('utf-8')
        text = font.Text(ft, txt, valign='center', halign='center')
        text.y = HEIGHT/2
        text.x = WIDTH/2

        x_margin = screen0.width - WIDTH

        @win.event
        def on_draw():
            win.clear()
            text.draw()

        win.set_location(x_margin, -HEIGHT)
        win.set_visible()
#        win.dispatch_events()
        def _fade_in():
#            print self._actual_infos
            for n in xrange(1, HEIGHT, 4):
                win.set_location(x_margin, -HEIGHT+n)
#                win.flip()
#                win.dispatch_events()
                yield
#            win.clear()
#            text.draw()
#            win.flip()
            DelayedAction(lambda: IterableAction(_fade_out()).start(0.01)).start(3)

        def _fade_out():
            for n in reversed(xrange(1, HEIGHT, 4)):
                win.set_location(x_margin, -HEIGHT+n)
#                win.dispatch_events()
                yield
            win.set_visible(False)
        IterableAction(_fade_in()).start(0.001)

def main():
    player_ctl = PlayerCtl()
    pp = PPlayer(player_ctl)
    player_ctl.views.append(pp)
    gtk.main()
    player_ctl.player.quit()

if __name__ == '__main__':
    main()



# vim: et ts=4 sw=4

def serve():
    startup('serve')

def startup(action='help'):
    import os
    import sys
    import zicbee.core.zshell as zshell
    if len(sys.argv) > 1:
        action = sys.argv[1]

    if '-' in sys.argv[0]:
        os.environ['ZDB'] = sys.argv[0].split('-', 1)[1]

    if action == 'use':
        os.environ['ZDB'] = sys.argv[2]
        del sys.argv[1:3] # Remove "use <db>"
        action = sys.argv[1]

    zshell.init()
    if '::' in action:
        params = action.split('::')
        action = params.pop(0)
        kparams = dict()
        for p in list(params):
            if '=' in p:
                k, v = p.split('=', 1)
                kparams[k] = v
                params.remove(p)
    else:
        kparams = dict()
        params = tuple()

    try:
        import zicbee.core.commands as cmds
        commands_dict = dict((i[3:], getattr(cmds, i)) for i in dir(cmds) if i.startswith('do_'))
        commands_dict.get(action, cmds.do_help)(*params, **kparams)
#        print dir(cmds)
#        exec('cmds.do_%s(*params, **kparams)'%action, commands_dict)
#        getattr(cmds, 'do_'+action, cmds.do_help)(*params, **kparams)
    except KeyboardInterrupt:
        print "Abort!"




from __future__ import with_statement

import pyglet
import pyglet.media.avbin # ensure avbin is installed
pyglet.options['audio'] = ('alsa', 'directsound', 'openal', 'silent')
from thread import start_new_thread
from time import sleep

from zicbee.core.zutils import DEBUG

#_eventdispatcher = pyglet.app.run
#start_new_thread(_eventdispatcher, tuple())

class SoundPlayer(object):
    def __init__(self):
        self._player = None
        self._source = None
        self.__initialized = False

    def __setup(self):
        start_new_thread(pyglet.app.run, tuple())
        self.__initialized = True

    def loadfile(self, filename, autoplay=True):
        if self._player:
            try:
                self._player.stop()
            except ValueError:
                pass # Stream not in play list
        try:
            self._source = pyglet.media.load(filename)
        except:
            DEBUG()
        else:
            if autoplay:
                self._player = self._source.play()

        if not self.__initialized:
            self.__setup()

    running = property(lambda self: self._player and self._player.playing)
    finished = property(lambda self: not bool(self._player._sources))

    def stop(self):
        self._player.stop()

    def volume(self, val):
        self._player.volume = val /100.0

    prop_volume = property(volume)

    def get_time_pos(self):
        return (self._source and self._source._packet.timestamp/1000000.0) or 0

    def seek(self, pos, t):
        self._source._seek(pos)

    def pause(self):
        if self._player.playing:
            self._player.pause()
        else:
            self._player.play()

    def quit(self):
        if self._player:
            self._player.stop()




#!/usr/bin/env python2.5
import os
import select
import subprocess

class_code = """# Access MPlayer from python
import os
import select
import subprocess

DEBUG='DEBUG' in os.environ

class MPlayer(object):
    ''' A class to access a slave mplayer process
    you may also want to use command(name, args*) directly

    Exemples:
        mp.command('loadfile', '/desktop/funny.mp3')
        mp.command('pause')
        mp.command('quit')

    Or:
        mp.loadfile('/desktop/funny.mp3')
        mp.pause()
        mp.quit()
    '''

    exe_name = 'mplayer' if os.sep == '/' else 'mplayer.exe'

    def __init__(self):
        self._spawn()

    def wait(self):
        self._mplayer.wait()

    def _spawn(self):
        self._mplayer = subprocess.Popen(
                [self.exe_name, '-cache', '128', '-slave', '-quiet', '-idle'],
                stdin=subprocess.PIPE, stdout=subprocess.PIPE, bufsize=1)
        self._readlines()

    def __del__(self):
        self._mplayer.stdin.write('quit\\n')

    def _readlines(self, timeout=0.4):
        ret = []
        while any(select.select([self._mplayer.stdout.fileno()], [], [], timeout)):
            ret.append( self._mplayer.stdout.readline() )
        return ret

    def _get_meta(self):
        try:
            meta = self.prop_metadata.split(',')
        except AttributeError:
            return None
        else:
            return dict(zip(meta[::2], meta[1::2]))

    meta = property(_get_meta, doc="Get metadatas as a dict"); del _get_meta

    def command(self, name, *args):
        ''' Very basic interface
        Sends command 'name' to process, with given args
        '''
        ret = self._readlines(0.01) # Flush
        if DEBUG:
            print "FLUSH LINES:", ret
        cmd = '%s%s%s\\n'%(name,
                ' ' if args else '',
                ' '.join(repr(a) for a in args)
                )
        if DEBUG:
            print "CMD:", cmd
        try:
            self._mplayer.stdin.write(cmd)
        except IOError:
            self._spawn()
            self._mplayer.stdin.write(cmd)

        if name == 'quit':
            return
        ret = self._readlines()
        if DEBUG:
            print "READ LINES:", ret

        if not ret:
            return None
        else:
            ret = ret[-1]

        if ret.startswith('ANS'):
            val = ret.split('=', 1)[1].rstrip()
            try:
                return eval(val)
            except:
                return val
        return ret

"""

exe_name = 'mplayer' if os.sep == '/' else 'mplayer.exe'
mplayer = subprocess.Popen([exe_name, '-input', 'cmdlist'],
        stdin=subprocess.PIPE, stdout=subprocess.PIPE)

def args_pprint(txt):
    lc = txt.lower()
    if lc[0] == '[':
        return '%s=None'%lc[1:-1]
    return lc

_slave_txt = None
def _find_doc_for(name):
    global _slave_txt
    if _slave_txt is None:
        import urllib
        uri = 'slave.txt' if os.path.isfile('slave.txt') else 'http://www.mplayerhq.hu/DOCS/tech/slave.txt'
        _slave_txt = urllib.urlopen(uri).readlines()

    actual_doc = None
    for line in _slave_txt:
        if actual_doc:
            if line.strip():
                actual_doc += line
            else:
                break
        elif line.endswith('properties:\n'):
                break
        elif line.startswith(name):
            actual_doc = line

    return actual_doc

while True:
    line = mplayer.stdout.readline()
    if not line:
        break
    if line[0].isupper():
        continue
    args = line.split()
    cmd_name = args.pop(0)
    arguments = ', '.join(args_pprint(a) for a in args)
    doc = _find_doc_for(cmd_name) or '%s(%s)'%(cmd_name, arguments)
    minargc = len([a for a in args if a[0] != '['])
    func_str = '''    def %(name)s(self, *args):
        """ %(doc)s
        """
        if %(condition)s:
            raise TypeError('%(name)s takes %(argc)d arguments (%%d given)'%%len(args))
        return self.command('%(name)s', *args)\n\n'''%dict(
            doc = doc,
            condition = ('len(args) != %d'%len(args)) if len(args) == minargc else ('not (%d <= len(args) <= %d)'%(minargc, len(args))),
            argc = len(args),
            name = cmd_name,
            )
    class_code += func_str

_not_properties = True
name = None
for line in _slave_txt:
    if _not_properties:
        if line.startswith('==================='):
            _not_properties = False
            class_code += '#Properties\n'
    else:
        if line[0].isalpha():
            if name is not None and name[-1] != '*':
                class_code += """
    prop_%(name)s = property(
        lambda self: self.get_property("%(name)s"),
        %(setter)s,
        doc = %(doc)s)
        """%dict(
                name = name,
                doc = ("'''%s'''"%('\n'.join(comments))) if any(comments) else 'None',
                setter = 'None' if ro else 'lambda self, val: self.set_property("%s", val)'%name
                )

            ridx = line.rindex('X')
            idx = line.index('X')
            try:
                name, p_type, limits = line[:idx].split(None, 2)
            except ValueError: # No limits
                name, p_type = line[:idx].split(None)
                limits = None
            permissions = len(line[idx:ridx+1])
            comments = [line[ridx+1:].strip()]

            ro = permissions < 2
        elif name is not None:
            comments.append(line.strip())

print class_code



# vim: et ts=4 sw=4

from zicbee.db import Database, DB_DIR
from zicbee.core.zshell import args, songs, DEFAULT_NAME
from zicbee.core.zutils import DEBUG
from zicbee.core.config import config
import urllib
import itertools

from .search import do_search
from .scan import do_scan
from .help import do_help
from .get import do_get
from .player import (do_play, do_pause,
        do_next, do_prev, do_shuffle,
        do_infos, do_playlist,
        do_tag, do_rate)


def do_kill(host=config.db_host):
    """ Kills the current db_host or any specified as argument """
    play_uri = 'http://%s/kill'%(host)
    try:
        urllib.urlopen(play_uri).read()
    except IOError:
        print "RIP."

def do_stfu(host=config.player_host):
    """ Kills the current player_host
    (in case db_host and player_host are the same, this command
    is equivalent to "kill")
    """
    play_uri = 'http://%s/kill'%(host)
    try:
        urllib.urlopen(play_uri).read()
    except IOError:
        print "RIP."

def do_serve(pure=False):
    """ Create a ZicDB instance
    parameters:
        pure (default: False): just start DB serving, no player
    """
    # chdir to serve files at the right place
    import os, sys
    from pkg_resources import resource_filename

    p = os.path.dirname(resource_filename('zicbee.ui.web', 'static'))
    os.chdir( p )

    # let's do webplayer
    import web
    from zicbee.player.webplayer import webplayer, web_db_index

    sys.argv = ['zicdb', '0.0.0.0:9090']
    try:
        import web.wsgiserver
        print "Running webplayer from", __file__
        if pure:
            urls = ('/db/(.*)', 'web_db_index',
                    '/(.*)', 'web_db_index')
        else:
            urls = ('/db/(.*)', 'web_db_index',
                    '/(.*)', 'webplayer')
        app = web.application(urls, locals())
        app.run()
    except:
        DEBUG()
        print os.kill(os.getpid(), 9)
        #print 'kill', os.getpid()

#do_serve.__doc__ = 
#def do_foo():
#    def _printall(*args):
#        print args
#    do_search(_printall)


def do_list():
    """ List available databases (some can be specified with "use" argument) """
    from os import listdir
    from os.path import isfile, join

    for i in listdir(DB_DIR):
        if isfile(join(DB_DIR, i, '__info__')) and isfile(join(DB_DIR, i, 'album')):
            txt = "%s # %d records"%(i, len(Database(i)))
            if i == DEFAULT_NAME:
                txt += ' [default]'
            print txt

def do_shell():
    """ Start a PDB (dev/hackers only)"""
    import pdb; pdb.set_trace()

def do_bundle():
    """ Dump used database to specified archive (any filename) """
    if len(args) != 1:
        raise SystemExit("Need filename name as agment !")
    songs.dump_archive(args[0])

def do_reset():
    """ Destroy all songs on used database """
    songs.destroy()
    print "Database cleared!"


def do_hash():
    """ List all songs by id and hash (mostly to debug find_dups command) """
    for i in songs.get_hash_iterator():
        print "%8d / %s"%i

def do_find_dups(wpt=None, ar=None):
    """
    Find duplicates (WIP)
    Parameters:
        wpt: wrong positive threshold (ceil to not reach), default == auto
        ar: auto remove (ask for directory deletion), the smallest directory always wins
    """
    import heapq
    from os.path import dirname

    hash_dict = dict()

    cnt = itertools.count()
    total_cnt = itertools.count()

    if wpt is None:
        wpt = min(1000, len(songs)/60) # take untaged/corrupted data into account

    for num, footprint in songs.get_hash_iterator():
        if footprint not in hash_dict:
            hash_dict[footprint] = [num]
        else:
            hash_dict[footprint].append(num)

    if ar:
        for m in (matches for num, matches in hash_dict.iteritems() if 1 < len(matches) < wpt):
            h = []
            for num in m:
                song = songs[num]
                heapq.heappush(h,
                        (len(song.filename), song))

            for nb, other in h:
                if other != h[-1][1]:
                    print "rm '%s'"%(other.filename.replace("'", r"\'"))
                else:
                    print "# kept %s"%other.filename
    else:
        for m in (matches for num, matches in hash_dict.iteritems() if 1 < len(matches) < wpt):
            print "#", cnt.next()
            for num in m:
                total_cnt.next()
                print "%d: %s"%(num, songs[num].filename)
        print total_cnt.next()-cnt.next()-1, "# songs to be removed..."


def do_fullhelp():
    """ The Hacker's help (WIP functions included) """
    import inspect
    g = globals()
    undoc = []
    command_functions = [g[name] for name in g.keys() if name[:3] == 'do_']
    command_functions.sort()
    commands_display = []
    remote_commands_display = []
    for cmd in command_functions:
        if cmd.__doc__:
            arg_names, not_used, neither, dflt_values = inspect.getargspec(cmd)
            if dflt_values is None:
                dflt_values = []
            else:
                dflt_values = list(dflt_values)

            # Ensure they have the same length
            if len(dflt_values) < len(arg_names):
                dflt_values = [None] * (len(dflt_values) - len(arg_names))
                map(None, arg_names, dflt_values)

            doc = ':'.join('%s%s'%(arg_name, '%s'%('='+str(arg_val) if arg_val is not None else '')) for arg_name, arg_val in itertools.imap(None, arg_names, dflt_values))

            if any(h for h in arg_names if h.startswith('host') or h.endswith('host')):
                out = remote_commands_display
            else:
                out = commands_display

            out.append( "%s[::%s]"%(cmd.func_name[3:], doc) if len(doc) > 1 else cmd.func_name[3:] ) # title
            out.append( "%s\n"%( ('\n'.join('   %s'%l for l in cmd.__doc__.split('\n') if l.strip()))) ) # body
        else:
            undoc.append(cmd.func_name[3:])

    for cmd in itertools.chain( ['[REMOTE COMMANDS]\n'], remote_commands_display, ['[LOCAL COMMANDS]\n'], commands_display ):
        print cmd
    print "Not documented:", ', '.join(undoc)




__all__ = ['Database', 'valid_tags']
# vim: ts=4 sw=4 et

import os
import buzhug
from zicbee.core.config import DB_DIR

valid_ext = ('.ogg','.mp3', '.mp4',
    '.aac', '.vqf', '.wmv', '.wma', '.m4a', '.asf', '.oga')

valid_tags = (
        'genre',
        'artist',
        'album',
        'title',
        'track',
        'filename',
        'score',
        'tags',
        'length')

filters_dict = dict(
        track = ('TRCK', 'tracknumber', u'WM/Track', 'trkn'),
        title = ('Title', 'TITLE', 'TIT1', 'TIT2', '\xa9nam'),
        artist = ('Author', 'author', 'AUTHOR', 'TPE1', '\xa9ART', '\xa9aART', 'WM/Composer', 'WM/Writer', 'TCOM', 'TOPE', 'album artist'),
        album = (u'WM/AlbumTitle', 'TALB', 'debut album'),
        genre = ('\xa9gen', u'WM/Genre'),
        # disc = ('disknumber',),
        # rating = ('Rating',),
        )


def checkdb(base_fn):
    return base_fn
    def _mkdec(somefn):
        def _auto_db_check(self, *args, **kw):
            if self.db is None:
                self._init()
            return somefn(self, *args, **kw)
        return _auto_db_check

    return _mkdec(base_fn)


class Database(object):

    DB_SCHEME = (
                ('filename', str),
                ('genre', unicode),
                ('artist', unicode),
                ('album', unicode),
                ('title', unicode),
                ('track', int),
                ('length', int),
                ('score', int),
                ('tags', unicode),
                )

    def __init__(self, name):
        """ Open/Create a database """
        self._db_dir = os.path.join(DB_DIR, name)
        self._init()

    def _init(self):
        self.db = buzhug.Base(self._db_dir)
        self.destroy = self.db.destroy
        self._open()

    @checkdb
    def __getitem__(self, item):
        return self.db[item]

    @checkdb
    def __len__(self):
        return len(self.db)

    @checkdb
    def search(self, *args, **kw):
        return self.db.select(*args, **kw)

    @checkdb
    def u_search(self, *args, **kw):
        return self.db.select_for_update(*args, **kw)

    @checkdb
    def _open(self, db=None):
        if db is None:
            db = self.db
        kw = dict(mode='open')
        db.create(*self.DB_SCHEME, **kw)

    @checkdb
    def dump_archive(self, filename):
        """ dump a bz2 archive from this database """
        self.db.cleanup()
        from tarfile import TarFile
        tar = TarFile.open(filename, 'w:bz2')
        prefix_sz = len(self._db_dir)

        for root, dirs, files in os.walk(self._db_dir):
            short_root = root[prefix_sz:]
            for fname in files:
                fd = file(os.path.join(root, fname))
                ti = tar.gettarinfo(arcname=os.path.join(short_root, fname), fileobj=fd)
                tar.addfile(ti, fileobj=fd)
        tar.close()

    def get_hash_iterator(self):
        """ Returns an (id, hash) tuple generator """
        for item in self.search():
            yield item.__id__, '%(artist)s:%(album)s:%(title)s'%dict(
                    (k, ''.join(c for c in getattr(item, k).lower() if c != ' '))
                    if isinstance(getattr(item, k), basestring)
                    else (k, getattr(item, k))
                    for k in item.fields)

    @checkdb
    def merge(self, directory=None, archive=None, no_dups=True):
        """ Merge informations from files in specified directory or archive """
        # TODO: add auto "no_dups" style after a len() check of the db

        try:
            EasyID3
        except NameError:
            from mutagen import File
            from mutagen.mp3 import MP3
            from mutagen.easyid3 import EasyID3

        us_prefix = None

        # Avoid duplicates
        if no_dups and len(self.db):

            # user specified prefix
            if directory is None:
                try:
                    print "Example filename %s"%(self.db[0].filename)
                except IndexError:
                    pass
                finally:
                    us_prefix = raw_input('Enter prefix to remove: ')
                    if us_prefix.strip():
                        directory = us_prefix

            # Remove every file starting with that directory
            old_len = len(self.db)
            for it in (i for i in self.db if i.filename.startswith(directory)):
                self.db.delete(it)
            print "Removed %d items"%( old_len - len(self.db) )
            self.db.cleanup()

        # Directory handling
        if directory is not None and not us_prefix:
            for root, dirs, files in os.walk(directory):
                for fname in files:
                    if fname[-4:].lower() in valid_ext:
                        fullpath = os.path.join(root, fname)

                        try:
                            tags = File(fullpath)
                        except Exception:
                            tags = None

                        if not tags:
                            yield "E"
                            continue
                        # Do it before tags is changed !
                        length = int(tags.info.length+0.5)
                        if isinstance(tags, MP3):
                            tags = EasyID3(fullpath)
                        data = filter_dict(dict(tags))
                        data['filename'] = fullpath
                        data['length'] = length
                        if data.get('title') and data.get('artist'):
                            yield '.'
                        else:
                            yield '0'
                        try:
                            self.db.insert(**data)
                        except:
                            import pdb; pdb.set_trace()

        # Archive handling
        if  archive is not None:
            from tarfile import TarFile
            from tempfile import mkdtemp
            from shutil import rmtree
            tar = TarFile.open(archive)
            tmp = mkdtemp()
            try:
                # Extract archive to disk
                buf_size = 2**16
                for ti in tar.getmembers():
                    out_fname = os.path.join(tmp, ti.name)
                    in_fd = tar.extractfile(ti)
                    out_fd = file(out_fname, 'w')
                    while True:
                        data = in_fd.read(buf_size)
                        if not data:
                            break
                        out_fd.write(data)

                    out_fd.close()

                # Do the buzhug part
                tmp_db = buzhug.Base(tmp)
                self._open(tmp_db)

                for alien_entry in tmp_db:
                    if alien_entry not in self.db:
                        entry_dict = dict()
                        for f in alien_entry.fields:
                            if f[0] == '_':
                                continue
                            val = getattr(alien_entry, f)
                            if isinstance(val, str) and f != 'filename':
                                val = unicode(val)
                            entry_dict[f] = val
                        self.db.insert(**entry_dict)
                        yield '.'
            finally:
                rmtree(tmp, ignore_errors=True)

        self.db.commit()

def filter_dict(data):
    """ Returns a filtered given data dict """
    for good_tag, bad_tags in filters_dict.iteritems():
        for bad_tag in bad_tags:
            if good_tag in data:
                break
            if data.get(bad_tag) is not None:
                data[good_tag] = data[bad_tag]

    for k, v in data.items():
        if k not in valid_tags:
            del data[k]
        else:
            while isinstance(v, (list, tuple)):
                try:
                    v = v[0]
                except IndexError:
                    v = None
            data[k] = v
    track_val = data.get('track')
    if track_val is not None:
        try:
            track_val = unicode(track_val).strip()
        except Exception, e:
            print e
            track_val = unicode(track_val.value)
        try:
            if isinstance(track_val, (list, tuple)):
                track_val = track_val[0]
            if not isinstance(track_val, int):
                track_val = int(track_val.replace('-', ' ').replace('/', ' ').split()[0])
        except:
            print "Unable to get track for", repr(track_val)
            data['track'] = None
        else:
            try:
                data['track'] = int(track_val.strip())
            except:
                data['track'] = None

    for k in ('genre', 'artist', 'album', 'title'):
        d = data.get(k)
        if d is None:
            data[k] = u''
        elif not isinstance(d, unicode):
            data[k] = unicode(d)

    if data.get('genre') == u'12':
        data['genre'] = u''

    return data




# vim: et ts=4 sw=4
import itertools
import os
import sys
from time import time
from zicbee.core.zshell import args, songs
from zicbee.core.zutils import duration_tidy, clean_path

def do_scan():
    """ Scan a directory for songs (fill Database)
    See "help" for a more complete documentation
    """
    if not args:
        sys.exit('At least one argument must be specified!')

    newline_iterator = itertools.cycle(x == 20 for x in xrange(21))
    orig_nb = len(songs)
    start_t = time()

    archives = []
    directories = []

    for path in args:
        path = clean_path(path)
        if os.path.isdir(path):
            directories.append(path)
        else:
            archives.append(path)

    def _scan(**kw):
        print ', '.join(':'.join((k,v)) for k,v in kw.iteritems())
        try:
            for status_char in songs.merge(**kw):
                print status_char,
                if newline_iterator.next():
                    print ''
                sys.stdout.flush()
        except Exception, e:
            print "ERROR!", str(e)
            import traceback
            traceback.print_exc()

    for path in archives:
        _scan(archive=path)

    for path in directories:
        _scan(directory=path)

    elapsed = time() - start_t
    delta = len(songs)-orig_nb
    print "\nProcessed %d (%s%d) songs in %s (%.2f/s.)"%(
            len(songs),
            '-' if delta < 0 else '+',
            delta,
            duration_tidy(elapsed),
            len(songs)/elapsed)


